/*
btdbck.c Version 0.1.0  Btree database check
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

int btdbck(btreefmt *bthdr)
   {
   int i;
   long ofst;
   int leafsw;
   int blknum;
   int adjadr;
   int prod;
   int currcd;
   int indx;
   int bit;
   int msk;
   int bitnum;
   int blks;
   int tot;
   int bitmapsz;
   unsigned char *p,*q;
   unsigned char *rcdptr;
   unsigned char *bitmap;
   bt_nodefmt *node;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   DWORD  filesize;
   bthdr->reason = 0;

   filesize = (DWORD) GetFileSize(bthdr->hndl, NULL);
   if (filesize == UINT_MAX)
      {
      msgerr = GetLastError();
      bthdr->reason = (int) SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btdbck: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btdbck: find first "
         "error on %s\n",
	 bthdr->dbname);
      bthdr->reason = SYS_ERR + msgerr;
      return(SYS_ERR + msgerr);
      } /* find first err */
   
   adjadr = (int) filesize - ROOT;
   blks   = adjadr / bthdr->blksz;
   prod   = blks   * bthdr->blksz;
   if (prod != adjadr)
      {
      fprintf(stderr,"btdbck: database size "
	 "not on record boundary\n");
      bthdr->reason = BTREE_ERR + 408;
      return(BTREE_ERR + 408);
      } /* if invalid record boundary */

   bitmapsz = (blks+7)>>3;
   bitmap = (unsigned char *) malloc(bitmapsz+16);
   if (bitmap == NULL)
      {
      fprintf(stderr,"btdbck: out of memory\n");
      bthdr->reason = BTREE_ERR + 401;
      return(BTREE_ERR + 401);
      } /* out of mem */
   p = (unsigned char *) bitmap;
   q = (unsigned char *) p + bitmapsz;
   while (p < q) *p++ = '\0';

   ofst = ROOT;
   tot = 0;
   while (tot < blks)
      {
      btgetblk(bthdr,ofst);
      if (bthdr->reason) return(bthdr->reason);
      if (bthdr->blk->right)
	 {
	 leafsw = 0;
         adjadr = bthdr->blk->right - ROOT;
         blknum = adjadr / bthdr->blksz;
         prod   = blknum * bthdr->blksz;
	 if (prod != adjadr)
            {
            fprintf(stderr,"btdbck: right pointer "
               "at %08x not on record boundary\n",
	       ofst);
            bthdr->reason = BTREE_ERR + 409;
            return(BTREE_ERR + 409);
            } /* if invalid record boundary */
	 if (blknum > blks)
            {
            fprintf(stderr,"btdbck: right pointer "
               "at %08x too large\n",
	       ofst);
            bthdr->reason = BTREE_ERR + 410;
            return(BTREE_ERR + 410);
            } /* if too large */
	 if (blknum < 1)
            {
            fprintf(stderr,"btdbck: right pointer "
               "at %08x too small\n",
	       ofst);
            bthdr->reason = BTREE_ERR + 411;
            return(BTREE_ERR + 411);
            } /* if too small */
         indx = blknum >> 3;
         bit  = blknum & 7;
         msk  = 1 << bit;
         p = (unsigned char *) bitmap + indx;
         if (*p & msk)
	    {
	    fprintf(stderr,"btdbck: "
	       "duplicate right pointer\n");
            bthdr->reason = BTREE_ERR + 402;
            return(BTREE_ERR + 402);
	    } /* if duplicate right pointer */
         *p |= msk;
	 } /* if index node */
      else
	 leafsw = 1;
      rcdptr = (unsigned char *)
	 bthdr->blk + bthdr->blkhdr;
      currcd = 1;
      node = (bt_nodefmt *) rcdptr;
      while (currcd <= bthdr->blk->count)
	 {
         node = (bt_nodefmt *) rcdptr;
	 if (!leafsw && !(node->left))
	    {
	    fprintf(stderr,"btdbck: missing "
	       "left pointer in index node\n");
            bthdr->reason = BTREE_ERR + 403;
            return(BTREE_ERR + 403);
	    } /* missing left in index node */
	 else if (leafsw && node->left)
	    {
	    fprintf(stderr,"btdbck: left "
	       "pointer in leaf node at %08x\n",
	       ofst);
            bthdr->reason = BTREE_ERR + 404;
            return(BTREE_ERR + 404);
	    } /* left pointer in leaf node */
	 if (!leafsw)
	    {
            adjadr = (node->left - ROOT);
            blknum = adjadr / bthdr->blksz;
            prod   = blknum * bthdr->blksz;
	    if (prod != adjadr)
               {
               fprintf(stderr,"btdbck: left pointer "
                  "at %08x record %d not on "
                  "record boundary\n",
	          ofst, currcd);
               bthdr->reason = BTREE_ERR + 412;
               return(BTREE_ERR + 412);
               } /* if invalid record boundary */
	    if (blknum > blks)
               {
               fprintf(stderr,"btdbck: left pointer "
                  "at %08x record %d too large\n",
	          ofst, currcd);
               bthdr->reason = BTREE_ERR + 413;
               return(BTREE_ERR + 413);
               } /* if too large */
	    if (blknum < 1)
               {
               fprintf(stderr,"btdbck: left pointer "
                  "at %08x record %d too small\n",
	          ofst, currcd);
               bthdr->reason = BTREE_ERR + 414;
               return(BTREE_ERR + 414);
               } /* if too small */
            indx = blknum >> 3;
            bit  = blknum & 7;
            msk  = 1 << bit;
            p = (unsigned char *) bitmap + indx;
            if (*p & msk)
	       {
	       fprintf(stderr,"btdbck: "
	          "duplicate left pointer\n");
               bthdr->reason = BTREE_ERR + 405;
               return(BTREE_ERR + 405);
	       } /* if duplicate left pointer */
            *p |= msk;
	    } /* if index node */
	 rcdptr = (unsigned char *)
	    rcdptr + bthdr->rcdnode;
	 currcd++;
	 } /* for each record in block */
      ofst += bthdr->blksz;
      tot++;
      } /* for each block in database */

   i = bitnum = 0;
   p = (unsigned char *) bitmap;
   while (i < bitmapsz)
      {
      bit = 1;
      while (bit < 255)
	 {
	 bitnum++;
	 if (bitnum == 1)
	    {
            if (*p & bit)
	       {
               fprintf(stderr,"btdbck: "
		  "byte %d bit %02x "
	          "root record %d\n",
	          i, bit, bitnum);
               bthdr->reason = BTREE_ERR + 406;
               return(BTREE_ERR + 406);
	       } /* invalid pointer */
	    } /* if not root block */
	 else if (bitnum <= blks)
	    {
            if (!(*p & bit))
	       {
               fprintf(stderr,"btdbck: "
		  "byte %d bit %02x "
	          "missing block %d\n",
	          i, bit, bitnum);
               bthdr->reason = BTREE_ERR + 407;
               return(BTREE_ERR + 407);
	       } /* missing overflow */
	    } /* if in non root block */
         bit <<= 1;
	 } /* for each bit in byte */
      p++;
      i++;
      } /* for each byte in bitmap */
   return(DBCKGOOD);
   } /* btdbck */
